#include "pch.h"
#include "Connection.h"
#include "Hook.h"
#include "ScatterGather.h"

//..............................................................................

int
Connection_create (
	Connection** resultConnection,
	Hook* hook,
	struct file* filp,
	const dm_IoctlDesc __user* ioctlDescArray_u,
	size_t ioctlDescCount,
	size_t pendingNotifySizeLimit,
	uint fileFlags
	)
{
	int result;
	const char* path;
	Connection* connection;
	size_t ioctlDescSize;
	dm_IoctlDesc* ioctlDescArray;
	HashTableEntry* entry;
	size_t i;

	connection = kmalloc (sizeof (Connection), GFP_KERNEL);
	if (!connection)
		return -ENOMEM;

	path = createPathString (&filp->f_path);
	if (IS_ERR (path))
	{
		kfree (connection);
		return PTR_ERR (path);
	}

	ioctlDescSize = sizeof (dm_IoctlDesc) * ioctlDescCount;
	ioctlDescArray = kmalloc (ioctlDescSize, GFP_KERNEL);
	if (!ioctlDescArray)
	{
		kfree (path);
		kfree (connection);
		return -ENOMEM;
	}

	result = copy_from_user (ioctlDescArray, ioctlDescArray_u, ioctlDescSize);
	if (result != 0)
	{
		kfree (ioctlDescArray);
		kfree (path);
		kfree (connection);
		return -EFAULT;
	}

	mutex_init (&connection->m_lock);
	INIT_LIST_HEAD (&connection->m_pendingReadList);
	INIT_LIST_HEAD (&connection->m_pendingNotifyList);
	init_waitqueue_head (&connection->m_notificationWaitQueue);
	HashTable_construct (&connection->m_ioctlDescMap, HashTableKeyType_Pointer, GFP_KERNEL);
	connection->m_ioctlDescArray = ioctlDescArray;
	connection->m_originalFilp = filp;
	connection->m_inode = filp->f_inode;
	connection->m_path = path;
	connection->m_hook = hook;
	connection->m_pendingReadCount = 0;
	connection->m_pendingNotifyCount = 0;
	connection->m_pendingNotifySize = 0;

	connection->m_pendingNotifySizeLimit = pendingNotifySizeLimit ?
		pendingNotifySizeLimit :
		dm_DefPendingNotifySizeLimit;

	connection->m_refCount = 1;
	connection->m_fileFlags = fileFlags;

	for (i = 0; i < ioctlDescCount; i++)
	{
		entry = HashTable_visit (
			&connection->m_ioctlDescMap,
			(const void*) (uintptr_t) ioctlDescArray [i].m_code
			);

		entry->m_value = &ioctlDescArray [i];
	}

	result = Hook_addConnection (hook, connection);
	if (result != 0)
	{
		kfree (ioctlDescArray);
		kfree (path);
		kfree (connection);
		return result;
	}

	*resultConnection = connection;
	return 0;
}

long
Connection_release (Connection* self)
{
	long refCount;

	refCount = atomicDec (&self->m_refCount);
	if (refCount)
		return refCount;

	Connection_disconnect (self);

	ASSERT (list_empty (&self->m_pendingReadList));
	ASSERT (list_empty (&self->m_pendingNotifyList));

	mutex_destroy (&self->m_lock);
	HashTable_destruct (&self->m_ioctlDescMap);
	kfree (self->m_ioctlDescArray);
	kfree (self->m_path);
	kfree (self);
	return 0;
}

void
Connection_disconnect (Connection* self)
{
	struct list_head* link;
	PendingRead* read;
	PendingNotify* notify;
	size_t notifySize;

	mutex_lock (&self->m_lock);

	while (!list_empty (&self->m_pendingReadList))
	{
		link = self->m_pendingReadList.next;
		list_del (link);
		read = container_of (link, PendingRead, m_link);
		self->m_pendingReadCount--;
		read->m_result = -ECANCELED;
		read->m_isCompleted = true;
		wake_up_interruptible (&read->m_waitQueue);
	}

	ASSERT (!self->m_pendingReadCount);
	self->m_pendingReadCount = 0;

	while (!list_empty (&self->m_pendingNotifyList))
	{
		link = self->m_pendingNotifyList.next;
		list_del (link);
		notify = container_of (link, PendingNotify, m_link);
		notifySize = sizeof (dm_NotifyHdr) + notify->m_notifyHdr.m_paramSize;

		self->m_pendingNotifyCount--;
		self->m_pendingNotifySize -= notifySize;
		kfree (notify);
	}

	ASSERT (!self->m_pendingNotifyCount && !self->m_pendingNotifySize);
	self->m_pendingNotifyCount = 0;
	self->m_pendingNotifySize = 0;

	mutex_unlock (&self->m_lock);

	Connection_detachFromHook (self);
}

void
Connection_detachFromHook (Connection* self)
{
	Hook* hook;

	mutex_lock (&self->m_lock);
	hook = self->m_hook;
	self->m_hook = NULL;
	mutex_unlock (&self->m_lock);

	if (hook)
		Hook_removeConnection (hook, self);
}

ssize_t
Connection_read (
	Connection* self,
	void __user* buffer_u,
	size_t size
	)
{
	int result;
	PendingRead read;
	PendingNotify* notify;
	dm_NotifyHdr notifyHdr;

	size_t notifySize;

	mutex_lock (&self->m_lock);

	if (!list_empty (&self->m_pendingNotifyList))
	{
		notify = container_of (self->m_pendingNotifyList.next, PendingNotify, m_link);
		notifySize = sizeof (dm_NotifyHdr) + notify->m_notifyHdr.m_paramSize;
		if (size < notifySize)
		{
			ASSERT (size >= sizeof (dm_NotifyHdr)); // should have been checked earlier

			notifyHdr = notify->m_notifyHdr;
			mutex_unlock (&self->m_lock);

			notifyHdr.m_flags |= dm_NotifyFlag_InsufficientBuffer;

			result = copy_to_user (buffer_u, &notifyHdr, sizeof (dm_NotifyHdr));
			if (result != 0)
				return -EFAULT;

			return sizeof (dm_NotifyHdr);
		}

		list_del (&notify->m_link);
		self->m_pendingNotifySize -= notifySize;
		self->m_pendingNotifyCount--;
		mutex_unlock (&self->m_lock);

		result = copy_to_user (buffer_u, &notify->m_notifyHdr, notifySize);

		kfree (notify);
		return result == 0 ? notifySize : -EFAULT;
	}

	if (self->m_fileFlags & O_NONBLOCK)
	{
		mutex_unlock (&self->m_lock);
		return -EWOULDBLOCK;
	}

	// TODO: use memory locking instead of double-buffering (user buffers may be HUGE!)

	read.m_buffer = kmalloc (size, GFP_KERNEL);
	if (!read.m_buffer)
	{
		mutex_unlock (&self->m_lock);
		return -ENOMEM;
	}

	init_waitqueue_head (&read.m_waitQueue);
	read.m_size = size;
	read.m_isCompleted = false;

	list_add_tail (&read.m_link, &self->m_pendingReadList);
	self->m_pendingReadCount++;
	mutex_unlock (&self->m_lock);

	result = wait_event_interruptible (read.m_waitQueue, read.m_isCompleted);
	if (result != 0)
	{
		kfree (read.m_buffer);
		return result;
	}

	if (read.m_result > 0)
	{
		result = copy_to_user (buffer_u, read.m_buffer, read.m_result);
		if (result != 0)
		{
			kfree (read.m_buffer);
			return -EFAULT;
		}
	}

	kfree (read.m_buffer);
	return read.m_result;
}

static
inline
void
Connection_p_markDataDropped_l (Connection* self)
{
	PendingNotify* notify;

	if (list_empty (&self->m_pendingNotifyList))
		return;

	notify = container_of (self->m_pendingNotifyList.prev, PendingNotify, m_link);
	notify->m_notifyHdr.m_flags |= dm_NotifyFlag_DataDropped;
}

void
Connection_notify (
	Connection* self,
	struct file* filp,
	uint16_t code,
	int result,
	uint32_t pid,
	uint32_t tid,
	uint64_t timestamp,
	MemBlock* paramBlockArray,
	size_t paramBlockCount
	)
{
	struct list_head* link;
	PendingRead* read;
	PendingNotify* notify;
	dm_NotifyHdr* notifyHdr;
	size_t paramSize;
	size_t notifySize;
	size_t actualSize;
	bool hasArgData;

	if (code == dm_NotifyCode_UnlockedIoctl || code == dm_NotifyCode_CompatIoctl)
	{
		hasArgData = Connection_p_preIoctlNotify (self, paramBlockArray, paramBlockCount);
		if (hasArgData)
			paramBlockCount++;
	}

	paramSize = getScatterGatherSize (paramBlockArray, paramBlockCount);
	notifySize = sizeof (dm_NotifyHdr) + paramSize;

	mutex_lock (&self->m_lock);
	if (filp == self->m_originalFilp) // don't dispatch close notification for the filp used to create this connection
	{
		if (code == dm_NotifyCode_Close)
			self->m_originalFilp = NULL;
		else
			printk (KERN_WARNING "tdevmon: unexpected notification on the original filp; code: %d\n", code); // flush? anyway, not a big deal

		mutex_unlock (&self->m_lock);
		return;
	}

	while (!list_empty (&self->m_pendingReadList))
	{
		link = self->m_pendingReadList.next;
		list_del (link);
		read = container_of (link, PendingRead, m_link);
		self->m_pendingReadCount--;
		mutex_unlock (&self->m_lock);

		ASSERT (read->m_size >= sizeof (dm_NotifyHdr)); // should have been checked

		notifyHdr = read->m_buffer;
		notifyHdr->m_signature = dm_NotifyHdrSignature;
		notifyHdr->m_code = code;
		notifyHdr->m_flags = 0;
		notifyHdr->m_result = result;
		notifyHdr->m_pid = pid;
		notifyHdr->m_tid = tid;
		notifyHdr->m_timestamp = timestamp;
		notifyHdr->m_paramSize = (uint32_t) paramSize;

		if (read->m_size >= notifySize)
		{
			copyScatterGather_u (notifyHdr + 1, paramBlockArray, paramBlockCount);
			actualSize = notifySize;
		}
		else
		{
			notifyHdr->m_flags |= dm_NotifyFlag_InsufficientBuffer;
			actualSize = sizeof (dm_NotifyHdr);
		}

		read->m_result = actualSize;
		read->m_isCompleted = true;
		wake_up_interruptible (&read->m_waitQueue);

		mutex_lock (&self->m_lock);

		if (actualSize == notifySize)
		{
			mutex_unlock (&self->m_lock);
			return;
		}
	}

	if (self->m_pendingNotifySize >= self->m_pendingNotifySizeLimit)
	{
		Connection_p_markDataDropped_l (self);
		mutex_unlock (&self->m_lock);
		return;
	}

	notify = kmalloc (sizeof (PendingNotify) + paramSize, GFP_KERNEL);
	if (!notify)
	{
		Connection_p_markDataDropped_l (self);
		mutex_unlock (&self->m_lock);
		return;
	}

	notifyHdr = &notify->m_notifyHdr;
	notifyHdr->m_signature = dm_NotifyHdrSignature;
	notifyHdr->m_code = code;
	notifyHdr->m_flags = 0;
	notifyHdr->m_result = result;
	notifyHdr->m_pid = pid;
	notifyHdr->m_tid = tid;
	notifyHdr->m_timestamp = timestamp;
	notifyHdr->m_paramSize = (uint32_t) paramSize;

	copyScatterGather_u (notifyHdr + 1, paramBlockArray, paramBlockCount);

	list_add_tail (&notify->m_link, &self->m_pendingNotifyList);
	self->m_pendingNotifyCount++;
	self->m_pendingNotifySize += notifySize;
	wake_up_interruptible (&self->m_notificationWaitQueue);
	mutex_unlock (&self->m_lock);
}

// . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

bool
Connection_p_preIoctlNotify (
	Connection* self,
	MemBlock* paramBlockArray,
	size_t paramBlockCount
	)
{
	dm_IoctlNotifyParams* notifyParams;
	const dm_IoctlDesc* ioctlDesc;
	size_t argSize;

	ASSERT (paramBlockCount == 1);

	notifyParams = (dm_IoctlNotifyParams*) paramBlockArray [0].m_p; // const cast

	ioctlDesc = HashTable_findValue (
		&self->m_ioctlDescMap,
		(const void*) (uintptr_t) notifyParams->m_code
		);

	if (!ioctlDesc)
	{
		notifyParams->m_argSize = 0;
		return false;
	}

	argSize = ioctlDesc->m_argFixedSize;

	if (ioctlDesc->m_flags & dm_IoctlFlag_HasArgSizeField)
	{
		// TODO: copy_from_user and add dynamic size from field
	}

	notifyParams->m_argSize = argSize;

	paramBlockArray [1].m_flags = MemBlockFlag_UserBuffer;
	paramBlockArray [1].m_p = (void*) (uintptr_t) notifyParams->m_arg;
	paramBlockArray [1].m_size = argSize;

	return true;
}

//..............................................................................
